/*=============================================================================================================
                                                   colony.cpp
===============================================================================================================

 Implementation of the class Colony
 
 C++-code accompanying: 
 
        authors : Boris H. Kramer, G. Sander van Doorn, Babak M. S. Arani, Ido Pen
        title : Eusociality and the evolution of aging in superorganisms
 

 Written by:
        Boris H. Kramer and G. Sander van Doorn
        Groningen Institute for Evolutionary Life Sciences (Gelifes)
        University of Groningen
        the Netherlands
 
 Program version
        01/03/2019	
 
 =============================================================================================================*/

// external dependencies 
#include <cassert>
#include <cmath>
#include "colony.h"
#include "random.h"

std::vector<Male> Colony::drones;

//default constructor for Colony
Colony::Colony(Queen * const q) : queen(q), fitness(0.0)
{

}

// destructor
Colony::~Colony() 
{
	if(queen != nullptr) delete queen;
	while (!workers.empty()) {
		delete workers.back();
		workers.pop_back();
	}
}

// reproduce function 
// includes a optimal life history model, where the queen and the workers make allocation decisions between foraging and egg laying;  foraging leads to exposure to different levels of extrinsic mortality given by the mortality functions; queens trade of resources vs reproduction
void Colony::reproduce()  
{
	double workerMortality;
	if (isQueenless()) { // workers only produce drones if there is no queen 
		const size_t n = workers.size();
		assert(n); //   check if n exist,and the program terminates if not true
		double optimalWorkerAllocation = 1.0 / (1.0 + sqrt(workerFecundity / productivity));// optimal worker allocation
		workerMortality = 1.0 - pow(1.0 - foragingRisk, 1.0 - optimalWorkerAllocation);// resulting worker mortality
		const double eggs = workerFecundity * optimalWorkerAllocation;// potential number off egs produced by a worker
		const double resources = productivity * (1.0 - optimalWorkerAllocation);//calculate resources obtained from foraging
		fitness = n * eggs * resources / (eggs + resources);// calculates fitness of the colony 
		for (size_t i = rnd::poisson(fitness); i > 0u; --i) 
			drones.push_back(Male(workers[rnd::integer(n)]));//add worker produced drones
	}
	else {  // if queenright
		double optimalQueenAllocation = size() / (1.0 + sqrt(queenFecundity / productivity)); // optimal queen allcoation, size() represents the size of the colony
		if (optimalQueenAllocation > 1.0) // limit optimal allocation to 1 
			optimalQueenAllocation = 1.0; //  only eggs produced  n extrinsic mortality
		queen->extrinsicMortality(1.0 - pow(1.0 - foragingRisk, 1.0 - optimalQueenAllocation)); //calculate extrinsic mortality for a queen thet forages / =(1-fR)^(1-oQA)  
		workerMortality = foragingRisk; // workers forage and are exposed to extrinsic mortality if the colony is queenright
		
		const double eggs = queenFecundity * optimalQueenAllocation;// number of eggs that the queen can produce if she doesnts stay in the colony and allocates time to foraging (at small colony sizes)
		const double resources = productivity * (1.0 - optimalQueenAllocation + workers.size());//resources retrieved from foraging
		fitness = eggs * resources / (eggs + resources); // if no resources eggs dont survive 
		for (size_t i = rnd::poisson(fitness); i > 0u; --i) { // get the actual number of  eggs produced in the time step
			Female *tmp = new Female(queen);//add new females
			tmp->extrinsicMortality(0.0); 
			offspring.push_back(tmp); // 
			drones.push_back(Male(queen)); //
		}
	}
	for (Female * worker : workers)
		worker->extrinsicMortality(workerMortality);// calculates worker survival probability 
}


// survival function 
void Colony::survival() 
{
	// worker survival
	for (size_t i = 0u; i < workers.size();) { // increase age of all workers
		workers[i]->incrementAge();
		if (workers[i]->doesSurvive()) ++i; // check if the individuals survives
		else {
			delete workers[i]; // remove workers that died
			workers[i] = workers.back();
			workers.pop_back();
		}
	}

	// queen survival 
	if (!isQueenless()) {// if queen is alive 
		queen->incrementAge();//increase age of queen
		if (!queen->doesSurvive()) {// if the queen doesnt survive
			delete queen;  //remove queen
			queen = nullptr;
		}
	}
	//polygynous scenario
	// queen replacement     
	while (isQueenless() && gyneReplacesQueen && offspring.size() && drones.size()) {// if  there is no queen and gyne replacement is true and there is offspring and drones in the colony then :
		queen = new Queen(offspring.back(), getDrone());// new queen queen is created from the last entry in the existing offspring vector (brood) of the colony
		if (!queen->doesSurvive()) { //then she has to survive
			delete queen;  // otherwise gets deleted 
			queen = nullptr;
		}
		delete offspring.back(); // when the queen was replaced by the offspring then  the last entry gets removed from the vector (the individuals that became the new queen 
		offspring.pop_back();
	}
	//colony inheritance
	if (isQueenless() && workerReplacesQueen && workers.size() && drones.size()) {// no queen , worker can replace queen = true ,workers in the colony 
		size_t i = rnd::integer(workers.size());//choose random worker
		queen = new Queen(workers[i], getDrone());//new queen mates with a male
		delete workers[i];//remove the worker
		workers[i] = workers.back();
		workers.pop_back();
	}
	//colony inheritance in case there are no workers but brood in the colony
	while (isQueenless() && workerReplacesQueen && offspring.size() && drones.size()) {
		queen = new Queen(offspring.back(), getDrone());// 
		if (!queen->doesSurvive()) { 
			delete queen;
			queen = nullptr;
		}
		delete offspring.back();
		offspring.pop_back();
	}
}

// function worker production
void Colony::addWorkers()
{
	while (!offspring.empty()) { //check if brood is in the colony
		if (offspring.back()->doesSurvive()) 
			workers.push_back(offspring.back());
		else
			delete offspring.back(); 
		offspring.pop_back();
	}
}

// select new queen for a colony
Female* Colony::getGyne()
{
	//std::cout << "offspring.size() =" << "\t" << offspring.size() << "\n";  // number of undeveloped offspring 
	assert(offspring.size());// test if the vector exists and has entries 
	Female *tmp = offspring.back();//get individual
	offspring.pop_back();//remove it from the brood
	return tmp;
}

// select male for a colony
const Male& Colony::getDrone() 
{
	assert(drones.size());//test if drones are available
	return drones[rnd::integer(drones.size())];// randomly choos one drone 
}

// function to write ouput file
void Colony::getData(size_t &sumn, size_t& sumq, size_t& sumw, size_t& suma, std::vector<double> &sumxq, std::vector<double> &sumxw) const 
{
	if (size()) {			 //if size is not empty
		++sumn;				 // number of occupied colonies
		if (!isQueenless()) {// if there is a queen 
			++sumq;			 //number of colonies with a queen
			suma += queen->getAge(); //get queen age
			queen->getData(sumxq); //obtain age specific survival for each individual
		}
		sumw += workers.size();//count workers
		for (Female * worker : workers)
			worker->getData(sumxw);//obtain age specific survival for each individual
	}
}