/*=============================================================================================================
===============================================================================================================

 Model implementation; entry point main()

 C++-code accompanying: 
 
        authors : Boris H. Kramer, G. Sander van Doorn, Babak M. S. Arani, Ido Pen
        title : Eusociality and the evolution of aging in superorganisms
 

 Written by:
        Boris H. Kramer and G. Sander van Doorn
        Groningen Institute for Evolutionary Life Sciences (Gelifes)
        University of Groningen
        the Netherlands
 
 Program version
        01/03/2019	

 ============================================================================================================*/

 //external dependencies
#include "colony.h"
#include "random.h"
#include "utils.h"
#include "parameters.h"
#include <numeric>
#include <algorithm>
#include <cassert>
#include <chrono>
#include <iterator>
#include <fstream>
#include <sstream>

 /*=============================================================================================================
													global variables
 =============================================================================================================*/

unsigned int simulationId;
Colony * population[nColony];//population objects, stores all the colonies
//for output
std::ofstream ofs;
std::ofstream ofParameters;


/*=============================================================================================================
												   model implementation
=============================================================================================================*/
//for putput
///Checks if a file exists for writing
bool is_regular_file(const std::string& filename)
{
	std::fstream f;
	f.open(filename.c_str(), std::ios::in);
	return f.is_open();
}

// construct  outfile names for parameters file 
std::string constructFilenameParameters(const long long simulationId)  
{
	std::stringstream par; // 
	par << "parameters_" << simulationId << ".csv";
	return par.str();   
}

//initialize the simulation
void init()
{
	//create simulation id and use it as a seed for the random number generator, all stored sim id are therefore alo the seeds for the random number generator. 
	simulationId = static_cast<unsigned int>(std::chrono::high_resolution_clock::now().time_since_epoch().count());
	rnd::set_seed(simulationId);
	// initialize mutational variance covariance matrix 
	Gene::initGeneticArchitecture();
	// create initial population with matd queens
	Female *q = new Female(); 
	for (int i = 0; i < nColony; ++i)
		population[i] = new Colony(new Queen(q, Male(q)));/
	delete q;

	// prepare data arrays
	for (int k = 0; k <= ageMax; ++k)
		Queen::ageOfDeath[k] = Female::ageOfDeath[k] = 0u; //initialize vectors for output
	
														   //prepare data output file 
	std::ostringstream str;
	str << "data_" << simulationId << ".csv";
	ofs.open(str.str());
	ofs << "t" << ',' << ',' // creates data_xx.csv file header
		<< "occupied" << ','
		<< "w/ queen" << ','
		<< "avg #workers" << ','
		<< "avg queen age" << ','
		<< "avg queen lifespan" << ','
		<< "avg worker lifespan" << ','
		<< ',' << "queen intrinsic survival";
	for (int a = 0; a < ageMax; ++a)
		ofs << ',' << "age class " << a + 2;
		ofs << ',' << "worker intrinsic survival";
	for (int a = 0; a < ageMax; ++a)
		ofs << ',' << "age class " << a+2; 
	ofs << "\n";

	// open and write parameters_xxx.csv  
	const std::string fileName = constructFilenameParameters(simulationId);
	ofParameters.open(fileName);														
	if (!ofParameters.is_open())														// test if file is open
		std::cout << "Something went wrong with opening the parameter file!" << "\n"; 	
		ofParameters << "Parameter" << "," << "Value" << "," << "Value" << "\n"			//header
		<< "nColony" << "," << nColony << "\n" 											
		<< "tEnd" << "," << tEnd << "\n"
		<< "dataInterval" << "," << dataInterval << "\n"
		<< "ageMax" << "," << ageMax << "\n"
		<< "foragingRisk" << "," << foragingRisk << "\n"
		<< "survivalOffset" << "," << survivalOffset << "\n"
		<< "productivity" << "," << productivity << "\n"
		<< "queenFecundity" << "," << queenFecundity << "\n"
		<< "workerFecundity" << "," << workerFecundity << "\n"
		<< "gyneReplacesQueen" << "," << gyneReplacesQueen << "\n"
		<< "workerReplacesQueen" << "," << workerReplacesQueen << "\n"
		<< "nGene" << "," << nGene << "\n"
		<< "nCaste" << "," << nCaste << "\n"
		<< "nTrait" << "," << nTrait << "\n"
		<< "mutationRate" << "," << mutationRate << "\n"
		<< "sigma[nCaste][nTrait]" << "," << *sigma[0] << "," << *sigma[1] << "\n"    
		<< "alpha[nTrait]" << "," << *alpha << "\n"           
		<< "beta[nTrait][(nCaste * (nCaste - 1)) / 2]" << "," << *beta[0] << "\n"
		<< " eta[nTrait]" << "," << *eta << "\n";     
	ofParameters.close();																// After you are done with the file always close it.
}

void dataAnalysis(int t) // function to generate output for the dataxxx.csv.
{
	std::vector<double> sumxq(ageMax, 0.0), sumxw(ageMax, 0.0);
	size_t sumw = 0u, sumq = 0u, sumn = 0u, suma = 0u;

	for (int i = 0; i < nColony; ++i)
		population[i]->getData(sumn, sumq, sumw, suma, sumxq, sumxw);//collects sumn=number of colonies ; sumq number of queens, sumw number of workers 
	
	ofs << t << ',' << ',' // output into data_simulationxxx.csv file 
		<< sumn * 1.0 / nColony << ','//occupied colonies
		<< sumq * 1.0 / sumn << ','   // colonies with queen
		<< sumw * 1.0 / sumn << ','   // average number of workers per colony
		<< (sumq ? suma * 1.0 / sumq : nan("")) << ',';

	sumn = suma = 0u; //clear the variables after writing to file 
	for (int k = 0; k <= ageMax; ++k) {//calculate mean queen lifespan
		double n = Queen::ageOfDeath[k];
		sumn += n;    //sums all the entries of the ages of death vector 
		suma += k * n;//
		Queen::ageOfDeath[k] = 0u; //clear variable after storing in file
	}
	ofs << suma * 1.0 / sumn << ','; // calculates average queen lifespan of the queens that have died in this timestep
	sumn = suma = 0u; //clear variable after storing in file

	for (int k = 0; k <= ageMax; ++k) {//calculate mean worker age
		double n = Female::ageOfDeath[k];
		sumn += n;
		suma += k * n;
		Female::ageOfDeath[k] = 0u;
	}
	ofs << suma * 1.0 / sumn << ',';

	for (int a = 0; a < ageMax; ++a)
		ofs << ',' << (sumq ? sumxq[a] / sumq : nan(""));// write mean age specific aurvival probability queen
	ofs << ',';
	for (int a = 0; a < ageMax; ++a)
		ofs << ',' << (sumw ? sumxw[a] / sumw : nan(""));// write mean age specific aurvival probability worker
	ofs << '\n';
	ofs.flush();
}

 //more output files create a sturcture for the measurements 
struct Measurement
{
	///Time of measurement, in timesteps
	int t;  // integer for timestep of the simulation

	///Queen ages. A negative number means NA, no queen
	std::vector<double> queenAges; //vector to store the queen ages 

	///Colony sizes
	std::vector<int> sizes; //vector to store the colony sizes

	///Fitmessess
	std::vector<double> fitnesses; // vector to store the fitnes of the colonies 
};

Measurement measure(const int t, Colony ** pop , const int nColonies)// member function
{
	Measurement m;
	m.t = t;
	//Measure all queen ages
	for (int i = 0; i != nColonies; ++i)
	{
		const Colony * const this_colony = pop[i];
		assert(this_colony);
		const Queen * const queen = this_colony->getQueen();//why this ?
		if (!queen)// no queen put -1 for the age
	{
			m.queenAges.push_back(-1.0);
		}
		else
		{
			m.queenAges.push_back(queen->getAge());// get the age and put it at the end of the vector
		}
	}
	//Measure colony sizes
	for (int i = 0; i != nColonies; ++i)
	{
		const Colony * const this_colony = pop[i];
		assert(this_colony);
		m.sizes.push_back(this_colony->size());
	}
	//Measure fitnesses
	for (int i = 0; i != nColonies; ++i)
	{
		const Colony * const this_colony = pop[i];
		assert(this_colony);
		m.fitnesses.push_back(this_colony->getFitness());
	}
	return m;
}



//write measurement into file 
std::ostream& operator<<(std::ostream& os, const Measurement& m) noexcept // writing to the outstream not sure why this has to be here? and when is it called? 
{
	assert(m.queenAges.size() == m.sizes.size());//checs if the size of the ages vector is similar to the sizes vector
	assert(m.queenAges.size() == m.fitnesses.size());

	const int sz = m.queenAges.size();
	for (int i = 0; i != sz; ++i)
	{
		os << m.t << ',' << m.queenAges[i] << ',' << m.sizes[i] << ',' << m.fitnesses[i] << '\n';// this basicall stores all the information i want to have in the results file at the end
	}
	return os; // this is all the data
}
//save measurements
void saveResults(const std::vector<Measurement>& results, const std::string& filename)
{
	std::ofstream f(filename);
	f << "t,queen_age,colony_size,fitness\n";

	for (const Measurement& m: results) f << m;
}

//write different ages at death file  deaths_time_xxx.csv
int writeFile(int t) {  
	std::ofstream aFile;
	aFile.open("deaths_time_"+std::to_string(t) +"_"+ std::to_string(simulationId) + ".csv");
	aFile << "t";
	for (int ha = 0; ha < ageMax; ++ha)
	{
		aFile << ',' << "queenAge_" << ha + 1;
	}
	for (int ha = 0; ha < ageMax; ++ha)
	{
		aFile << ',' << "workerAge_" << ha + 1;
	}
	aFile << "\n";
	aFile << t;
	for (int ha = 0; ha < ageMax; ++ha)
	{
		aFile << "," << Queen::ageOfDeath[ha];
	}
	for (int ha = 0; ha < ageMax; ++ha)
	{
		aFile << "," << Female::ageOfDeath[ha];
	}
	aFile << "\n";
	aFile.close();
	return 0;
}


/*=============================================================================================================
												   main()
=============================================================================================================*/
int main()    // the designated start of the program
{
	//preliminaries
	init();
	std::vector<Measurement> results;
	// for each timestep: 
	for (int t = 0; t < tEnd; ) {
		// production of offspring and survival
		size_t sum = 0u;
		std::vector<size_t> weights(nColony, 0u); 
		for (int i = 0; i < nColony; ++i) {
			if (population[i]->size()) { // if there are individuals in the colony 
				population[i]->reproduce();//reproduce 
				population[i]->survival(); //survive?
				sum += weights[i] = population[i]->nrOffspring(); // count the number of offspring per colony and sum it all up 
			}
		}

		// establishment of new colonies throughout the simulation only when the colony is emppty (no workers, no queen and no offspring (weights) 
		for (int i = 0; i < nColony && sum; ) { // go through all colonies
			if (population[i]->size() == 0u && weights[i] == 0u) {  // i = colony we are looking at j colony the gyne is taken from // 02022017 added : && weights[i] == 0u to make sure no individual are taken from the empty offspring vector 
				size_t j = std::discrete_distribution<size_t>(weights.begin(), weights.end())(rnd::rng); // to calculate potential fitness ..expected based on fitness
				Female * tmp = population[j]->getGyne(); // get new queen
				Queen * q = new Queen(tmp, Colony::getDrone());
				delete tmp;// delete temp female
				if (q->doesSurvive()) {
					delete population[i]; 
					population[i] = new Colony(q);// new colony established
					++i;
				}
				else delete q;
				--weights[j];
				--sum;
			}
			else ++i;
		}

		bool terminate = true;
		for (int i = 0; i < nColony; ++i) {
			population[i]->addWorkers();
			if (population[i]->size())
				terminate = false;
		}
		Colony::clearDrones();

		if (terminate) {
			std::cout << "population extinct\n";
			break;
		}
		
		if (t == tEnd - 1 || t == tEnd - 2)//output the age ofdeath vector to a file 
		{
			writeFile(t);// stores the age at death distribution but just one line so far 
		}

		++t;//update the simulation time after everything is done in the timestep

		if (t % dataInterval == 0)  // run analysis at each dataInterval
		{
			dataAnalysis(t);

		}  
		if (t == tEnd )// at the end of simulation time 
		{
			const Measurement result = measure(t, population, nColony); //collect data
			results.push_back(result);// adds the measures for the next timesteps 
		}
		
	}

	//saving the  queenAge, colony size and fitness of the colony to a file after the simulation ends
	const std::string results_filename = "results_" + std::to_string(simulationId) + ".csv";// create filename for the size age results 
	saveResults(results, results_filename); // here the results file is saved (end of simulation)
	assert(is_regular_file(results_filename));//check if there is a regular file 

	for (int i = 0; i < nColony; ++i)
		delete population[i];//delete all colonies after the end of the simulation
	return 0;
}

